geo

package
v0.4.4 Latest Latest
Warning

This package is not in the latest version of its module.

Go to latest
Published: Aug 2, 2024 License: MIT Imports: 2 Imported by: 0

Documentation

Index

Constants

This section is empty.

Variables

This section is empty.

Functions

func NewBoundsAround

func NewBoundsAround(point primitives.Point, radius float64) (*primitives.Rect, error)

NewBoundsAroundPoint creates a new bounding rectangle given a center point, and a distance from the center point in meters.

Types

This section is empty.

Jump to