Documentation
¶
Index ¶
Constants ¶
This section is empty.
Variables ¶
This section is empty.
Functions ¶
func NewBoundsAround ¶
func NewBoundsAround(point primitives.Point, radius float64) (*primitives.Rect, error)
NewBoundsAroundPoint creates a new bounding rectangle given a center point, and a distance from the center point in meters.
Types ¶
This section is empty.
Click to show internal directories.
Click to hide internal directories.